"""
Module for statistical helper methods.
Power methods are received from DataCamp Customer Analytics and AB Testing Course
"""

from scipy import stats


def get_power(n, p1, p2, cl):
    """Returns the power of our hypothesis test by number of samples,
    probabilities (conversion_rate) and confidence level

    Parameters
    ----------
    n : int : Number of samples
        
    p1 : float: probability for first set
        
    p2 : float: probability for second set
        
    cl : float: confidence level
        

    Returns
    -------
    Power : float:

    """
    alpha = 1 - cl

    qu = stats.norm.ppf(1 - alpha / 2)

    diff = abs(p2 - p1)
    bp = (p1 + p2) / 2

    v1 = p1 * (1 - p1)
    v2 = p2 * (1 - p2)

    bv = bp * (1 - bp)

    power_part_one = stats.norm.cdf((n ** 0.5 * diff - qu * (2 * bv) ** 0.5) / (v1 + v2) ** 0.5)
    power_part_two = 1 - stats.norm.cdf((n ** 0.5 * diff + qu * (2 * bv) ** 0.5) / (v1 + v2) ** 0.5)

    power = power_part_one + power_part_two
    return power


def get_sample_size(power, p1, p2, cl, max_n=1000000):
    """calculates the power for different test size until it reaches the desired power level,
    and returns the test size

    Parameters
    ----------
    power : int:
        
    p1 : float: probability for first set
        
    p2 : float: probability for second set
        
    cl : float: confidence level
        
    max_n : int: maximum number of tests to be run
         (Default value = 1000000)

    Returns
    -------
    n : int: sample size
    """
    n = 1
    while n <= max_n:
        tmp_power = get_power(n, p1, p2, cl)
        if tmp_power >= power:
            return n
        else:
            n = n + 1



"""
Module for analysis of python functions
"""

import pandas
import time
from functools import wraps


def timer(func):
    """Calculate and print duration of function

    Parameters
    ----------
    func : Function:
        

    Returns
    -------

    """
    @wraps(func)
    def wrapper(*args, **kwargs):
        """

        Parameters
        ----------
        *args : Arguments
            
        **kwargs : Keyword Arguments
            

        Returns
        -------
        Output: output of inner function.
        """
        start = pandas.Timestamp('today')
        val = func(*args, **kwargs)
        end = pandas.Timestamp('today')
        delta = end - start
        print(f'Process took: {delta.components.minutes} min, {delta.components.seconds},'
              f'secs {delta.components.milliseconds} ms')
        return val
    return wrapper


def return_type(func):
    """Print the return type of inner function

    Parameters
    ----------
    func : Function:
        

    Returns
    -------

    """
    @wraps(func)
    def wrapper(*args, **kwargs):
        """

        Parameters
        ----------
        *args : Arguments
            
        **kwargs : Keyword arguments
            

        Returns
        -------
        Output: output of inner function.
        """
        val = func(*args, **kwargs)
        if not val:
            print(f'Process does not return any value - {type(val)}')
        elif len([val]) > 1:
            print([type(n) for n in val])
        else:
            print(type(val))
        return val
    return wrapper


# if __name__ == '__main__':
#
#     @timer
#     def wait(name):
#         time.sleep(2)
#         print(f'Hello {name}')
#
#     wait('Levent')
#     #
#     # @return_type
#     # def return_func(num1, num2):
#     #     return num1*30
#     #
#     # print(return_func(1, 3))
#
#     # @run_n_times(3)
#     # def greet(name):
#     #     print(f'Hello {name}')
#     #
#     # greet('Levent')



"""
Module for handling files, directories and databases
"""

import contextlib
import os
import sqlite3


@contextlib.contextmanager
def working_directory(path):
    """Change the working directory during process by context manager

    Parameters
    ----------
    path : str: Path for target directory
        

    Returns
    -------
    None
    """
    cwd = os.getcwd()
    os.chdir(path)
    yield None
    os.chdir(cwd)


@contextlib.contextmanager
def sqlite_database(path):
    """Connect to sqlite database from path and disconnect after operation using context manager

    Parameters
    ----------
    path : Path to sqlite database file
        

    Yields
    -------
    conn: sqlite.Connection : Connection object for db operations
    """
    conn = None
    try:
        conn = sqlite3.connect(path)
        yield conn
    finally:
        conn.close()
        print('DB Disconnected')


# if __name__ == '__main__':
    # print(f'Before {os.listdir()}')
    # with working_directory('../'):
    #     print(f'During {os.listdir()}')
    # print(f'After {os.listdir()}')

    # with sqlite_database('trackerdb.db') as conn:
    #     sql = f"SELECT * FROM kpi_data"
    #     cursor = conn.cursor()
    #     cursor.execute(sql)
    #     try:
    #         obj = cursor.fetchall()
    #         print(obj)
    #     except sqlite3.Error as err:
    #         print('Error:', err)



"""
Module for python helper methods
"""

from functools import wraps


def run_n_times(n):
    """Returns a decorator to run decorated function n number of times

    Parameters
    ----------
    n : int : Number of times to run function
        

    Returns
    -------

    """
    def decorator(func):
        """

        Parameters
        ----------
        func : function:
            

        Returns
        -------

        """
        def wrapper(*args, **kwargs):
            """

            Parameters
            ----------
            *args : Arguments of wrapped function
                
            **kwargs : Keyword arguments of wrapped function
                

            Returns
            -------
            None
            """
            for i in range(n):
                func(*args, **kwargs)
        return wrapper
    return decorator


def memoize(func):
    """Store the results of the decorated function for fast lookup

    Parameters
    ----------
    func : Function:
        

    Returns
    -------

    """

    cache = {}

    @wraps(func)
    def wrapper(*args, **kwargs):
        """

        Parameters
        ----------
        *args : Arguments of wrapped function
            
        **kwargs : Keyword arguments of wrapped function
            

        Returns
        -------
        Output: output of internal function or cache
        """
        # If these arguments haven't been seen before,
        if (args, str(kwargs)) not in cache.keys():
            print('not in cache')
            print(cache)
            # Call func() and store the result.
            cache[(args, str(kwargs))] = func(*args, **kwargs)
        return cache[(args, str(kwargs))]
    return wrapper


# if __name__ == '__main__':
    # @memoize
    # def testfunc(num1, num2, name, age):
    #     time.sleep(3)
    #     print(f'Running operations for {name}')
    #     return num1*num2+age
    #
    # res = testfunc(1, 2, name='levent', age=39)
    # print(res)
    # res = testfunc(1, 3, name='elif', age=39)
    # print(res)
    # res = testfunc(1, 2, name='levent', age=39)
    # print(res)
    # res = testfunc(1, 3, name='elif', age=39)
    # print(res)
    # res = testfunc(1, 3, name='fatma', age=39)
    # print(res)
    # res = testfunc(1, 2, name='levent', age=39)
    # print(res)

    # @timer
    # def wait(name):
    #     time.sleep(1.5)
    #     print(f'Hello {name}')
    # wait('Levent')
    #
    # @return_type
    # def return_func(num1, num2):
    #     return num1*30
    #
    # print(return_func(1, 3))

    # @run_n_times(3)
    # def greet(name):
    #     print(f'Hello {name}')
    #
    # greet('Levent')



"""
Module for model evaluation methods
"""

from sklearn.metrics import accuracy_score, f1_score
from sklearn.metrics import classification_report
from sklearn.metrics import confusion_matrix
import pandas


def classification_score(y_test, y_pred):
    """Method for printing evaluation results for binary classification

    Parameters
    ----------
    y_test : array or pandas.Series:
        Array of actual labels/classes
        
    y_pred : array or pandas.Series:
        Array of predictions

    Returns
    -------
    None :
        Prints the f1 score, accuracy score, classification report and confusion matrix
    """
    f1 = f1_score(y_test, y_pred)
    accuracy = accuracy_score(y_test, y_pred)
    print(f'Accuracy Score = {accuracy}\nF1 Score:{f1}\n')
    print(classification_report(y_test, y_pred))
    print(pandas.DataFrame(confusion_matrix(y_test, y_pred)[::-1, ::-1],
                           columns=['Pred_True', 'Pred_False'],
                           index=['Act_True', 'Act_False']))



"""
Levent Tools for Python
------------------------
leventools is a package for miscellaneous tools for data science and python scripts
"""
from leventools import data_science
from leventools import tools



from setuptools import setup, find_packages

setup(
    author='Levent Soykan',
    description='Package of helper tools for data science and python',
    name='leventools',
    version='0.1.1',
    packages=find_packages(include=['leventools', 'leventools.*']),
    python_requires='>=3.6',
    install_requires=[
        'numpy>=1.23'
        'pandas>=1.5'
        'scikit-learn>=1.1'
        'scipy>=1.9'
    ]
)



"""
Module for Exploratory Data Analysis methods
"""

import pandas
import numpy
from scipy import stats


def identify_outliers(dataframe: pandas.DataFrame, column: str):
    """Identify and return outliers based on specific column.
    Cutoff : IQR*1.5 is used

    Parameters
    ----------
    dataframe: pandas.DataFrame :
    Input DataFrame
        
    column: str :
    Column for outlier detection

    Returns
    -------
    outliers: List :
    List of outliers based on input column

    lower: float :
    Lower boundary for outliers, values below this are outliers

    upper: float :
    Upper boundary for outliers, values above this are outliers
    """
    # calculate inter quartile range
    q25, q75 = numpy.percentile(dataframe[column], 25), numpy.percentile(dataframe[column], 75)
    iqr = q75 - q25
    print(f'COLUMN: {column}\nPercentiles: 25th={q25:.2f}, 75th={q75:.2f}, IQR={iqr:.2f}')
    # calculate the outlier cutoff
    cut_off = iqr * 1.5
    lower, upper = q25 - cut_off, q75 + cut_off
    # identify outliers
    outliers = [x for x in dataframe[column] if x < lower or x > upper]
    print(f'Identified outliers: {len(outliers)} of {len(dataframe[column])} values')
    # return outliers and boundaries
    return outliers, lower, upper


def remove_outliers(dataframe: pandas.DataFrame, column: str):
    """Identify and remove outliers based on specific column

    Parameters
    ----------
    dataframe: pandas.DataFrame :
    Input DataFrame
        
    column: str :
    Column for outlier detection

    Returns
    -------
    dataframe: pandas.DataFrame :
    Output DataFrame with outliers removed

    """
    _, lower, upper = identify_outliers(dataframe, column)
    return dataframe[(dataframe[column] >= lower) & (dataframe[column] <= upper)]


def chi2_correlation(dataframe: pandas.DataFrame, column1: str, column2: str):
    """Check correlation between 2 categorical variables using chi2 test

    Parameters
    ----------
    dataframe: pandas.DataFrame :
    Input DataFrame
        
    column1: str :
        
    column2: str :

    Returns
    -------
    p-Value: float:
        p_value result of Chi2 test

    contingency_table: pandas.DataFrame:
        table showing counts among columns

    """
    contingency_table = pandas.crosstab(dataframe[column1], dataframe[column2], margins=False)
    stat, p, dof, expected = stats.chi2_contingency(contingency_table)
    return p, contingency_table
